@@ -1,5 +1,8 @@
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE DefaultSignatures #-}
+#if WITH_CALLSTACK
+{-# LANGUAGE ImplicitParams #-}
+#endif
 #if WITH_TEMPLATE_HASKELL
 {-# LANGUAGE TemplateHaskell #-}
 #endif
@@ -36,6 +39,8 @@ module Control.Monad.Logger
     , LoggingT (..)
     , runStderrLoggingT
     , runStdoutLoggingT
+    , runChanLoggingT
+    , unChanLoggingT
     , withChannelLogger
     , filterLogger
     , NoLoggingT (..)
@@ -46,6 +51,12 @@ module Control.Monad.Logger
     , logWarn
     , logError
     , logOther
+    -- * TH logging of showable values
+    , logDebugSH
+    , logInfoSH
+    , logWarnSH
+    , logErrorSH
+    , logOtherSH
     -- * TH logging with source
     , logDebugS
     , logInfoS
@@ -67,7 +78,14 @@ module Control.Monad.Logger
     , logWarnNS
     , logErrorNS
     , logOtherNS
-
+#if WITH_CALLSTACK
+    -- * Callstack logging
+    , logDebugCS
+    , logInfoCS
+    , logWarnCS
+    , logErrorCS
+    , logOtherCS
+#endif
     -- * utilities for defining your own loggers
     , defaultLogStr
     , Loc (..)
@@ -80,10 +98,11 @@ import Language.Haskell.TH.Syntax (Lift (lift), Q, Exp, Loc (..), qLocation)
 import Data.Monoid (Monoid)
 
 import Control.Applicative (Applicative (..))
+import Control.Concurrent.Chan (Chan(),writeChan,readChan)
 import Control.Concurrent.STM
 import Control.Concurrent.STM.TBChan
 import Control.Exception.Lifted (onException)
-import Control.Monad (liftM, ap, when, void)
+import Control.Monad (liftM, ap, when, void, forever)
 import Control.Monad.Base (MonadBase (liftBase))
 import Control.Monad.Loops (untilM)
 import Control.Monad.Trans.Control (MonadBaseControl (..), MonadTransControl (..))
@@ -133,6 +152,10 @@ import Control.Monad.Reader.Class ( MonadReader (..) )
 import Control.Monad.State.Class  ( MonadState (..) )
 import Control.Monad.Writer.Class ( MonadWriter (..) )
 
+#if WITH_CALLSTACK
+import GHC.Stack as GHC
+#endif
+
 import Prelude hiding (catch)
 
 #if MIN_VERSION_fast_logger(2, 1, 0)
@@ -243,7 +266,16 @@ instance (MonadLoggerIO m, Monoid w) => MonadLoggerIO (Strict.RWST r w s m)
 #if WITH_TEMPLATE_HASKELL
 logTH :: LogLevel -> Q Exp
 logTH level =
-    [|monadLoggerLog $(qLocation >>= liftLoc) (pack "") $(lift level) . (id :: Text -> Text)|]
+    [|monadLoggerLog $(qLocation >>= liftLoc) (pack "") $(lift level)
+     . (id :: Text -> Text)|]
+
+-- | Generates a function that takes a 'LogLevel' and a 'Show a => a'.
+--
+-- @since 0.3.18
+logTHShow :: LogLevel -> Q Exp
+logTHShow level =
+    [|monadLoggerLog $(qLocation >>= liftLoc) (pack "") $(lift level)
+      . ((pack . show) :: Show a => a -> Text)|]
 
 -- | Generates a function that takes a 'Text' and logs a 'LevelDebug' message. Usage:
 --
@@ -267,6 +299,31 @@ logError = logTH LevelError
 logOther :: Text -> Q Exp
 logOther = logTH . LevelOther
 
+
+-- | Generates a function that takes a 'Show a => a' and logs a 'LevelDebug' message. Usage:
+--
+-- > $(logDebugSH) (Just "This is a debug log message")
+--
+-- @since 0.3.18
+logDebugSH :: Q Exp
+logDebugSH = logTHShow LevelDebug
+
+-- | See 'logDebugSH'
+logInfoSH :: Q Exp
+logInfoSH = logTHShow LevelInfo
+-- | See 'logDebugSH'
+logWarnSH :: Q Exp
+logWarnSH = logTHShow LevelWarn
+-- | See 'logDebugSH'
+logErrorSH :: Q Exp
+logErrorSH = logTHShow LevelError
+
+-- | Generates a function that takes a 'Show a => a' and logs a 'LevelOther' message. Usage:
+--
+-- > $(logOtherSH "My new level") "This is a log message"
+logOtherSH :: Text -> Q Exp
+logOtherSH = logTHShow . LevelOther
+
 -- | Lift a location into an Exp.
 --
 -- Since 0.3.1
@@ -579,6 +636,29 @@ runStderrLoggingT = (`runLoggingT` defaultOutput stderr)
 runStdoutLoggingT :: MonadIO m => LoggingT m a -> m a
 runStdoutLoggingT = (`runLoggingT` defaultOutput stdout)
 
+-- | Run a block using a @MonadLogger@ instance which writes tuples to an
+--   unbounded channel.
+--
+--   The tuples can be extracted (ie. in another thread) with `unChanLoggingT`
+--   or a custom extraction funtion, and written to a destination.
+--
+-- @since 0.3.17
+runChanLoggingT :: MonadIO m => Chan (Loc, LogSource, LogLevel, LogStr) -> LoggingT m a -> m a
+runChanLoggingT chan = (`runLoggingT` sink chan)
+    where
+        sink chan loc src lvl msg = writeChan chan (loc,src,lvl,msg)
+
+-- | Read logging tuples from an unbounded channel and log them into a
+--   `MonadLoggerIO` monad, forever.
+--
+--   For use in a dedicated thread with a channel fed by `runChanLoggingT`.
+--
+-- @since 0.3.17
+unChanLoggingT :: (MonadLogger m, MonadIO m) => Chan (Loc, LogSource, LogLevel, LogStr) -> m ()
+unChanLoggingT chan = forever $ do
+    (loc,src,lvl,msg) <- liftIO $ readChan chan
+    monadLoggerLog loc src lvl msg
+
 -- | Within the 'LoggingT' monad, capture all log messages to a bounded
 --   channel of the indicated size, and only actually log them if there is an
 --   exception.
@@ -692,3 +772,65 @@ logErrorNS src = logWithoutLoc src LevelError
 
 logOtherNS :: MonadLogger m => Text -> LogLevel -> Text -> m ()
 logOtherNS = logWithoutLoc
+
+#if WITH_CALLSTACK
+-- Callstack based logging
+
+mkLoggerLoc :: GHC.SrcLoc -> Loc
+mkLoggerLoc loc =
+  Loc { loc_filename = GHC.srcLocFile loc
+      , loc_package  = GHC.srcLocPackage loc
+      , loc_module   = GHC.srcLocModule loc
+      , loc_start    = ( GHC.srcLocStartLine loc
+                       , GHC.srcLocStartCol loc)
+      , loc_end      = ( GHC.srcLocEndLine loc
+                       , GHC.srcLocEndCol loc)
+      }
+
+locFromCS :: GHC.CallStack -> Loc
+locFromCS cs = case getCallStack cs of
+                 ((_, loc):_) -> mkLoggerLoc loc
+                 _            -> defaultLoc
+
+logCS :: (MonadLogger m, ToLogStr msg)
+      => GHC.CallStack
+      -> LogSource
+      -> LogLevel
+      -> msg
+      -> m ()
+logCS cs src lvl msg =
+  monadLoggerLog (locFromCS cs) src lvl msg
+
+-- | Logs a message with location given by 'CallStack'.
+-- See 'Control.Monad.Logger.CallStack' for more convenient
+-- functions for 'CallStack' based logging.
+--
+-- @since 0.3.19
+logDebugCS :: MonadLogger m => GHC.CallStack -> Text -> m ()
+logDebugCS cs msg = logCS cs "" LevelDebug msg
+
+-- | See 'logDebugCS'
+--
+-- @since 0.3.19
+logInfoCS :: MonadLogger m => GHC.CallStack -> Text -> m ()
+logInfoCS cs msg = logCS cs "" LevelInfo msg
+
+-- | See 'logDebugCS'
+--
+-- @since 0.3.19
+logWarnCS :: MonadLogger m => GHC.CallStack -> Text -> m ()
+logWarnCS cs msg = logCS cs "" LevelWarn msg
+
+-- | See 'logDebugCS'
+--
+-- @since 0.3.19
+logErrorCS :: MonadLogger m => GHC.CallStack -> Text -> m ()
+logErrorCS cs msg = logCS cs "" LevelError msg
+
+-- | See 'logDebugCS'
+--
+-- @since 0.3.19
+logOtherCS :: MonadLogger m => GHC.CallStack -> LogLevel -> Text -> m ()
+logOtherCS cs lvl msg = logCS cs "" lvl msg
+
+#endif
